"""
IA Parc Job Handler
"""
import uuid
from iaparc_inference.config import LOGGER
from .listener import IAPListener


class IAPJobHandler:
    def __init__(self, 
                 inputs: str, 
                 outputs: str,
                 decode=True, 
                 config_path: str = "/opt/pipeline/pipeline.json"):
        """
        Constructor for IAPJobHandler
        Arguments:
        - inputs:    Job input name
        - outputs:   Job output name
        Optional Arguments:
        - decode:    Whether to decode the output
        - config_path: Path to the pipeline configuration file        
        """
        self.uuid = str(uuid.uuid4())
        self.inputs = inputs + "." + self.uuid
        self.outputs = outputs + "." + self.uuid
        self.listener = IAPListener(
            self.handle_reply, 
            decode=decode, 
            inputs=self.outputs, 
            outputs=self.inputs, 
            config_path=config_path)
        if inputs not in self.listener.inputs:
            LOGGER.error(f"Input {inputs} not found in pipeline")
            raise ValueError(f"Input {inputs} not found in pipeline")
        if outputs not in self.listener.outputs:
            LOGGER.error(f"Output {outputs} not found in pipeline")
            raise ValueError(f"Output {outputs} not found in pipeline")
        input_item = self.listener.inputs[inputs]
        self.listener.error_queue = self.listener.queue + "." + input_item["link"] 


    # def handle_reply(self, data):
        


"""
IA Parc Inference data handler
"""
import os
import io
import logging
import logging.config
from typing import Any
import iaparc_inference.decoders as decoders

Error = ValueError | None

LEVEL = os.environ.get('LOG_LEVEL', 'INFO').upper()
logging.basicConfig(
    level=LEVEL,
    force=True,
    format="%(name)s: %(message)s",
    datefmt="%Y-%m-%d %H:%M:%S",
)
LOGGER = logging.getLogger("Inference")
LOGGER.propagate = True


def decode(raw: bytes, content_type: str, conf: dict) -> tuple[Any, Error]:
    if conf["type"] == "multimodal":
        raw_items, error = decoders.decode_multipart(
            raw, conf["items"], content_type)
        result = {}
        for item in conf["items"]:
            item_data = raw_items.get(item["name"])
            if item_data:
                result[item["name"]], error = _decode(item_data, item["type"])
                if error:
                    LOGGER.error(f"Error decoding {item['name']}: {error}")
                    return None, error
        return result, None
    else:
        return _decode(raw, conf["type"])

def _decode(raw, kind) -> tuple[Any, Error]:
    """
    Decode data
    """
    match kind:
        case "integer":
            return decoders.decode_int(raw)
        case "boolean":
            return decoders.decode_bool(raw)
        case "number":
            return decoders.decode_float(raw)
        case "file" | "binary" | "audio" | "video":
            return decoders.decode_file(raw)
        case "text" | "string":
            return decoders.decode_text(raw)
        case "image":
            return decoders.decode_image(raw)
        case "json" | "matrix" | "array":
            return decoders.decode_json(raw)
        case _:
            return raw, ValueError(f"Unsupported data type: {kind}")




"""
IA Parc Inference decoders
"""
from struct import unpack
from PIL.Image import Image
from io import BytesIO

Error = ValueError | None

## Data decoders

def decode_float(data: bytes) -> tuple[float, Error]:
    """
    Read float
    Arguments:
    data: bytes
    """
    if data is None:
        return 0.0, ValueError("No data to read")
    if not isinstance(data, bytes):
        return 0.0, ValueError("Data is not bytes")
    try:
        value = unpack('f', data)[0]
    except Exception as e:
        return 0.0, ValueError(f"Error reading float: {e}")
    return value, None

def decode_int(data: bytes) -> tuple[int, Error]:
    """
    Read int
    Arguments:
    data: bytes
    """
    if data is None:
        return 0, ValueError("No data to read")
    if not isinstance(data, bytes):
        return 0, ValueError("Data is not bytes")
    try:
        value = unpack('i', data)[0]
    except Exception as e:
        return 0, ValueError(f"Error reading int: {e}")
    return value, None

def decode_bool(data: bytes) -> tuple[bool, Error]:
    """
    Read bool
    Arguments:
    data: bytes
    """
    if data is None:
        return False, ValueError("No data to read")
    if not isinstance(data, bytes):
        return False, ValueError("Data is not bytes")
    try:
        value = unpack('?', data)[0]
    except Exception as e:
        return False, ValueError(f"Error reading bool: {e}")
    return value, None

def decode_file(data: bytes) -> tuple[BytesIO, Error]:
    """
    Read file
    Arguments:
    data: bytes
    """
    if not data:
        return BytesIO(), ValueError("No data to read")
    if not isinstance(data, bytes):
        return BytesIO(), ValueError("Data is not bytes")
    try:
        file = BytesIO(data)
    except Exception as e:
        return BytesIO(), ValueError(f"Error reading file: {e}")
    return file, None

def decode_image(data: bytes) -> tuple[Image|None, Error]:
    """
    Read image
    Arguments:
    data: bytes
    """
    if not data:
        return None, ValueError("No data to read")
    if not isinstance(data, bytes):
        return None, ValueError("Data is not bytes")
    try:
        from PIL import Image
        image = Image.open(BytesIO(data))
    except Exception as e:
        return None, ValueError(f"Error reading image: {e}")    
    return image, None

def decode_text(data: bytes) -> tuple[str, Error]:
    """
    Read text
    Arguments:
    data: bytes
    """
    if not data:
        return "", ValueError("No data to read")
    if not isinstance(data, bytes):
        return "", ValueError("Data is not bytes")
    try:
        text = data.decode("utf-8")
    except Exception as e:
        return "", ValueError(f"Error reading text: {e}")
    return text, None

def decode_json(data: bytes) -> tuple[dict, Error]:
    """
    Read json
    Arguments:
    data: bytes
    """
    if not data:
        return {}, ValueError("No data to read")
    if not isinstance(data, bytes):
        return {}, ValueError("Data is not bytes")
    try:
        from json_tricks import loads
        json_data = loads(data.decode("utf-8"))
    except Exception as e:
        return {}, ValueError(f"Error reading json: {e}")
    return json_data, None

def decode_numpy(data: bytes) -> tuple[dict, Error]:
    """
    Read numpy
    Arguments:
    data: bytes
    """
    return decode_json(data)

def decode_multipart(data: bytes, items: list, content_type: str) -> tuple[dict, Error]:
    """
    Read multi-part data
    Arguments:
    data: bytes
    """
    if not data:
        return {}, ValueError("No data to read")
    if not isinstance(data, bytes):
        return {}, ValueError("Data is not bytes")
    if not items:
        return {}, ValueError("No items to read")
    try:        
        from streaming_form_data import StreamingFormDataParser
        from streaming_form_data.targets import BaseTarget

        class BytesTarget(BaseTarget):
            def __init__(self, *args, **kwargs):
                super().__init__(*args, **kwargs)
                self._data = None
                
            def on_data_received(self, data: bytes):
                self.data = data

        res = {}
        results = {}
        if "boundary" not in content_type:
            boundary = _get_boundary(data)
            if boundary:
                content_type += f'; boundary={boundary}'
        
        headers = {'Content-Type': content_type}
        #headers = {'Content-Type': 'multipart/form-data; boundary=boundary'}
        parser = StreamingFormDataParser(headers=headers)
        for item in items:
            res[item["name"]] = BytesTarget()                
            parser.register(item["name"], res[item["name"]])
        parser.data_received(data)
        
        for item in items:
            results[item["name"]] = res[item["name"]].data

    except Exception as e:
        return {}, ValueError(f"Error reading multi-part data: {e}")
    return results, None

def decode_audio(data: bytes):
    """
    Read audio
    Arguments:
    data: bytes
    """
    pass


def _get_boundary(data: bytes) -> str | None:
    """
    Get boundary
    Arguments:
    data: bytes
    """
    splitted = data.split(b"\r\n")
    if len(splitted) < 2:
        return None
    boundary = splitted[0]
    if len(boundary) < 2:
        return None
    return boundary[2:].decode("utf-8")


from __future__ import annotations

from ast import Subscript
import asyncio
from typing import (
    TYPE_CHECKING,
    AsyncIterator,
    Awaitable,
    Callable,
    List,
    Optional,
)
from uuid import uuid4

from nats.errors import TimeoutError, ConnectionClosedError
from nats.aio.msg import Msg
from nats.aio.subscription import Subscription

class BatchSubscription:

    def __init__(self, sub: Subscription, batch_size: int):
        self._sub = sub
        self._batch_size = batch_size
        self._dones = {}
        self._pending_tasks = {}
    
    async def get_batch(self, timeout: Optional[float] = 0.005) -> List[Msg]:
        msgs: List[Msg] = []
        task_name = str(uuid4())
        if timeout is None:
            timeout = 0.005
        self._dones[task_name] = False
        try:
            msg = await self._sub.next_msg(timeout=600)
            msgs.append(msg)
            future = asyncio.create_task(
                asyncio.wait_for(self._wait_for_next_batch(task_name, timeout), timeout)
            )
            self._pending_tasks[task_name] = future
            await future
        except asyncio.TimeoutError:
            if self._sub._conn.is_closed:
                raise ConnectionClosedError
            self._dones[task_name] = True
            raise TimeoutError
        except asyncio.CancelledError:
            if self._sub._conn.is_closed:
                raise ConnectionClosedError
            self._dones[task_name] = True
            raise
        else:
            msgs.extend(self._pending_tasks[task_name])
            return msgs
        finally:
            self._pending_tasks.pop(task_name, None)
            self._dones.pop(task_name, None)
        
        return msgs
    
    async def _wait_for_next_batch(self, task_name: str, timeout: float = 0.005) :
        self._pending_tasks[task_name] = []
        limit = self._batch_size - 1
        wait_time = timeout / 5
        if wait_time < 0.005:
            wait_time = 0.001
        while not self._dones[task_name]:
            try:
                #msg = await self._sub._pending_queue.get()
                msg = await self._sub.next_msg(timeout=wait_time)
                self._sub._pending_size -= len(msg.data)
                #self._sub._pending_queue.task_done()
                self._pending_tasks[task_name].append(msg)
                if len(self._pending_tasks[task_name]) >= limit:
                    break
            except asyncio.CancelledError:
                break
            except TimeoutError:
                break
            except Exception as e:
                print("Unknown exception:", e)
                break


"""
IA Parc Inference service
Support for inference of IA Parc models
"""
from json import dumps
from math import e
import os
import time
import asyncio
from re import T
import time
import uuid
from inspect import signature
import logging
import logging.config
import nats
from iaparc_inference.config import Config
from iaparc_inference.data_decoder import decode
from iaparc_inference.data_encoder import DataEncoder
from iaparc_inference.subscription import BatchSubscription

Error = ValueError | None

LEVEL = os.environ.get('LOG_LEVEL', 'INFO').upper()
logging.basicConfig(
    level=LEVEL,
    force=True,
    format="%(name)s: %(message)s",
    datefmt="%Y-%m-%d %H:%M:%S",
)
LOGGER = logging.getLogger("Inference")
LOGGER.propagate = True


class IAPListener():
    """
    Inference Listener class
    """

    def __init__(self,
                 callback,
                 decode=False,
                 batch:int=-1,
                 inputs:str = "",
                 outputs:str = "",
                 config_path:str= "/opt/pipeline/pipeline.json",
                 url:str="",
                 queue:str=""
                 ):
        """
        Constructor
        Arguments:
        - callback:     Callback function to proccess data
                        callback(data: Any | list[Any], parameters: Optional[dict])
        Optional arguments:
        - inputs:       Input queue name
        - outputs:      Output queue name
        - decode:       Set wether data should be decoded before calling the callback function (default: True)
        - batch:        Batch size for inference (default: -1)
                        If your model do not support batched input, set batch to 1
                        If set to -1, batch size will be determined by the BATCH_SIZE
                        environment variable
        - config_path:  Path to config file (default: /opt/pipeline/pipeline.json)
        - url:          Url of inference server (default: None)
                        By default determined by the NATS_URL environment variable,
                        however you can orverride it here
        - queue:        Name of queue (default: None)
                        By default determined by the NATS_QUEUE environment variable,
                        however you can orverride it here
        """
        # Init internal variables
        self.decode = decode
        self.timeout = 0.002
        self.exec_time = 0
        self._subs_in = []
        self._subs_out = []
        self._dag = Config(config_path)
        if inputs:
            self._dag.inputs = inputs
        self._links_in = self._dag.inputs.split(",")
        self._links_out = self._dag.outputs.split(",")

        self.lock = asyncio.Lock()
        self.callback = callback
        sig = signature(callback)
        self.callback_args = sig.parameters
        nb_params = len(self.callback_args)
        if nb_params == 1:
            self.callback_has_parameters = False
        else:
            self.callback_has_parameters = True

        if url:
            self.url = url
        else:
            self.url = os.environ.get("NATS_URL", "nats://localhost:4222")
        if queue:
            self.queue = queue.replace("/", "-")
        else:
            self.queue = os.environ.get(
                "NATS_QUEUE", "inference").replace("/", "-")
        if batch > 0:
            self.batch = batch
        else:
            self.batch = int(os.environ.get("BATCH_SIZE", 1))
        if self.batch > 1:
            self.is_batch = True
        else:
            self.is_batch = False

        self.error_queue = self.queue + ".ERROR"
        self.inputs = {}
        self.outputs = {}
        self.encoders = {}
        if len(self._dag.pipeline) == 0:
            print("No pipeline defined")
            quit(1)
        # Get inputs from first entity
        entity = self._dag.pipeline[0]
        for item in entity.input_def:
            if "link" in item and item["link"] in self._links_in:
                self.inputs[item["link"]] = item
        
        # Get outputs from last entity
        entity = self._dag.pipeline[-1]
        for item in entity.output_def:
            if "link" in item and item["link"] in self._links_out:
                if "name" in item:
                    item_name = item["name"]
                else:
                    item_name = item["link"]
                self.outputs[item_name] = item["link"]
                self.encoders[item["link"]] = DataEncoder(item)
        if outputs and outputs in self.outputs:
            self.default_output = self.outputs[outputs]
        else:
            self.default_output = self._links_out[0]

    @property
    def dag(self) -> Config:
        """ Input property """
        return self._dag

    @property
    def inputs_name(self) -> list:
        """ Input property """
        return self._links_in

    def run(self):
        """
        Run inference service
        """
        asyncio.run(self._run_async())

    async def _run_async(self):
        """ Start listening to NATS messages
        url: NATS server url
        batch_size: batch size
        """
        self.nc = await nats.connect(self.url)
        self.js = self.nc.jetstream()

        for q_name in self.inputs_name:
            #item = self.inputs[q_name]
            queue_in = self.queue + "." + q_name
            print("Listening on queue:", queue_in)
            js_in = await self.js.subscribe(queue_in+".>",
                                            queue=self.queue+"-"+q_name,
                                            stream=self.queue)
            self._subs_in.append((q_name, js_in))
            nc_in = await self.nc.subscribe("nc."+queue_in+".*.*", queue=self.queue+"-"+q_name)
            self._subs_in.append((q_name, nc_in))

        print("Default queue out:", self.default_output)
        self.data_store = await self.js.object_store(bucket=self.queue+"-data")

        os.system("touch /tmp/running")
        tasks = []
        for name, sub_in in self._subs_in:
            tasks.append(self.wait_msg(name, sub_in))
        await asyncio.gather(*tasks)
        await self.nc.close()

    async def wait_msg(self, name, sub_in):
        # Fetch and ack messagess from consumer.
        if sub_in.subject[:7] == "_INBOX.":
            subject = sub_in.subject[7:]
            is_js = True
        else:
            subject = sub_in.subject.replace(".*.*", "")
            is_js = False
        if self.is_batch:
            batch_sub = BatchSubscription(sub_in, self.batch)
        while True:
            if not self.is_batch:
                try:
                    msg = await sub_in.next_msg(timeout=600)
                except nats.errors.TimeoutError:
                    continue
                except TimeoutError:
                    continue
                except nats.errors.ConnectionClosedError:
                    LOGGER.error(
                        "Fatal error message handler: ConnectionClosedError")
                    break
                except asyncio.CancelledError:
                    LOGGER.error(
                        "Fatal error message handler: CancelledError")
                    break
                except Exception as e: # pylint: disable=W0703
                    LOGGER.error("Unknown error:", exc_info=True)
                    continue
                
                # Message received
                await asyncio.gather(
                    self.handle_msg(subject, name, [msg]),
                    self.term_msg([msg], is_js)
                )
            else:
                msgs = []
                ok = False
                try:
                    msgs = await batch_sub.get_batch(self.timeout)
                except nats.errors.TimeoutError:
                    continue
                except TimeoutError:
                    continue
                except nats.errors.ConnectionClosedError:
                    LOGGER.error(
                        "Fatal error message handler: ConnectionClosedError")
                    break
                except asyncio.CancelledError:
                    LOGGER.error(
                        "Fatal error message handler: CancelledError")
                    break
                
                # Messages received
                t0 = time.time()
                await asyncio.gather(
                    self.handle_msg(subject, name, msgs),
                    self.term_msg(msgs, is_js)
                )
                t1 = time.time()
                if self.exec_time == 0:
                    self.exec_time = t1 - t0
                self.exec_time = (self.exec_time + t1 - t0) / 2
                if self.exec_time < 0.02:
                    self.timeout = 0.002
                elif self.exec_time > 0.35:
                    self.timeout = 0.05
                else:
                    self.timeout = self.exec_time * 0.15

    async def handle_msg(self, subject, name, msgs):
        async with self.lock:
            if self.is_batch:
                uids, sources, batch, params_lst, content_types = zip(*[await self.get_data(subject, msg) for msg in msgs])
                batch = list(batch)
                await self._process_data(name, uids, sources, batch, content_types, params_lst)
            else:
                for msg in msgs:
                    uid, source, data, params, content_type = await self.get_data(subject, msg)
                    await self._process_data(name, [uid], [source], [data], [content_type], [params])

        return

    async def term_msg(self, msgs, is_js=False):
        if is_js:
            for msg in msgs:
                await msg.ack()
        else:
            ack = "".encode("utf8")
            for msg in msgs:
                await msg.respond(ack)

    async def get_data(self, subject, msg):
        l_sub = len(subject) + 1
        uid = msg.subject[(l_sub):]
        source = msg.headers.get("DataSource", "")
        params_lst = msg.headers.get("Parameters", "")
        params = {}
        if params_lst:
            for p in params_lst.split(","):
                args = p.split("=")
                if len(args) == 2:
                    k, v = args
                    params[k] = v
        content_type = msg.headers.get("ContentType", "")
        data = None
        if source == "object_store":
            obj_res = await self.data_store.get(msg.data.decode())
            data = obj_res.data
        else:
            data = msg.data

        return (uid, source, data, params, content_type)

    async def send_msg(self, out, uid, source, data, parameters={}, error=""):
        if error is None:
            error = ""
        _params = dumps(parameters)
        breply = "".encode()
        contentType = ""
        if out != self.error_queue:
            _out = self.queue + "." + out + "." + uid
            #print("Sending reply to:", _out)
            if data is not None:
                if source == "object_store":
                    store_uid = str(uuid.uuid4())
                    breply = store_uid.encode()
                    err = None
                    if isinstance(data, (bytes, bytearray)):
                        bdata = data
                    else:
                        bdata, contentType, err = self.encoders[out].encode(data)
                        if err:
                            _out = self.error_queue + "." + uid
                            breply = str(err).encode()
                            error = "Error encoding data"
                    if not err:
                        await self.data_store.put(store_uid, bdata)
                else :
                    if isinstance(data, (bytes, bytearray)):
                        breply = data
                    else:
                        breply, contentType, err = self.encoders[out].encode(data)
                        if err:
                            _out = self.error_queue + "." + uid
                            breply = str(err).encode()
                            error = "Error encoding data"
                    if len(breply) > 8388608: # 8MB
                        store_uid = str(uuid.uuid4())
                        source = "object_store"
                        bdata = breply
                        breply = store_uid.encode()
                        await self.data_store.put(store_uid, bdata)
        else:
            _out = self.error_queue + "." + uid
            breply = data.encode()
        
        headers = {"ProcessError": error,
                   "ContentType": contentType,
                   "DataSource": source,
                   "Parameters": _params}
        
        if out != self.error_queue:
            try:
                nc_out = "nc." + _out
                await self.nc.request(nc_out, breply, timeout=60, headers=headers)
                _sent = True
            except nats.errors.NoRespondersError:
                await self.js.publish(_out, breply, headers=headers)
            except Exception as e: # pylint: disable=W0703
                LOGGER.error("Error sending message on core NATS:", exc_info=True)
        else:
            await self.js.publish(_out, breply, headers=headers)

    async def _process_data(self, name: str,
                      uids: list,
                      sources: list,
                      raw_datas: list,
                      content_types: list,
                      reqs_parameters: list):
        """
        Process data
        Arguments:
        - requests:   list of data to process
        - is_batch:   is batched data
        """
        LOGGER.debug("handle request")
        queue_out = self.default_output
        p_datas = []
        p_sources = []
        p_uids = []
        p_params = []
        for uid, src, raw, ctype, params in zip(uids, sources, raw_datas, content_types, reqs_parameters):
            if self.decode:
                data, error = decode(raw, ctype, self.inputs[name])
                if error:
                    asyncio.create_task(self.send_msg(self.error_queue,
                                                      uid,
                                                      src,
                                                      str(error),
                                                      params,
                                                      "Wrong input"))
                    continue
                p_datas.append(data)
            else:
                p_datas.append(raw)
            p_sources.append(src)
            p_uids.append(uid)
            p_params.append(params)
        
        try_error = ""
        if len(p_datas) > 0:
            try:
                error = ""
                if self.is_batch:
                    if self.callback_has_parameters:
                        res = self.callback(p_datas, p_params)
                    else:
                        res = self.callback(p_datas)
                    if isinstance(res, tuple):
                        if len(res) == 2:
                            result, error = res
                        if len(res) == 3:
                            result, out, error = res
                            if out in self.outputs:
                                queue_out = self.outputs[out]
                    else:
                        result = res
                    if not isinstance(result, list):
                        error = "batch reply is not a list"
                    if len(p_datas) != len(result):
                        error = "batch reply has wrong size"
                    if error:
                        for uid, source, params in zip(p_uids, p_sources, p_params):
                            asyncio.create_task(self.send_msg(queue_out,
                                                              uid,
                                                              source,
                                                              error,
                                                              params,
                                                              error))
                    else:
                        for uid, source, res, params in zip(p_uids, p_sources, result, p_params):
                            asyncio.create_task(self.send_msg(queue_out,
                                                              uid,
                                                              source,
                                                              res,
                                                              params))
                else:
                    if len(p_params) > 0:
                        _params = p_params[0]
                    else:
                        _params = {}
                    if self.callback_has_parameters:
                        res = self.callback(p_datas[0], _params)
                    else:
                        res = self.callback(p_datas[0])
                    if isinstance(res, tuple):
                        if len(res) == 2:
                            result, error = res
                        if len(res) == 3:
                            result, out, error = res
                            if out in self.outputs:
                                queue_out = self.outputs[out]
                    else:
                        result = res
                    asyncio.create_task(self.send_msg(queue_out,
                                                      p_uids[0],
                                                      p_sources[0],
                                                      result,
                                                      _params,
                                                      error=error))
                
            except ValueError:
                LOGGER.error("Fatal error message handler", exc_info=True)
                try_error  = "Wrong input"
            except Exception as e: # pylint: disable=W0703
                LOGGER.error("Fatal error message handler", exc_info=True)
                try_error = f'Fatal error: {str(e)}'
            if try_error:
                for uid, source in zip(p_uids, p_sources):
                    asyncio.create_task(self.send_msg(
                        self.error_queue, uid, src, try_error, "Wrong input"))




"""
IA Parc Inference data decoder
"""
import operator
from functools import reduce  # forward compatibility for Python 3
from typing import Any
import iaparc_inference.encoders as encoders

Error = ValueError | None


class DataEncoder():
    """
    Data Encoder
    This a read-only class that encodes the data
    """

    def __init__(self, conf: dict,):
        """
        Constructor
        Arguments:
        
        """
        self.conf = conf
        self.name = conf["name"]
        self.encoders = {}
        if "type" in self.conf and self.conf["type"] == "multimodal":
            for item in self.conf["items"]:
                self.encoders[item["name"]] = DataEncoder(item)
        if "type" in self.conf and self.conf["type"] == "json":
            self.json_images, self.json_types = self.check_json_images(self.conf["items"])
        else:
            self.json_images = []
            self.json_types = []
    
    def check_json_images(self, items: list) -> tuple[list, list]:
        elts = []
        types = []
        for item in items:
            item_name = item.get("name")
            if item_name:
                if item.get("type") in ["image", "file", "binary", "audio", "video"]:
                    elts.append([item.get("name")])
                    types.append(item.get("type"))
                elif item.get("type") == "json":
                    res, _types = self.check_json_images(item["items"])
                    if len(res) > 0:
                        for k, t in zip(res, _types):
                            new_elt = [item.get("name")]
                            new_elt.extend(k)
                            elts.append(new_elt)
                            types.append(t)
        return elts, types
    
    def encode(self, data: Any) -> tuple[bytes, str, Error]:
        """
        Decode data
        Arguments:
        data: Any
        """
        res = ''.encode()
        contentType = ""
        err = None
        if data is None:
            return res, "", ValueError("No data to encode")
        try:
            match self.conf["type"]:
                case "multimodal":
                    form_data = {}
                    if not isinstance(data, dict):
                        return res, "", ValueError("Data is not a dictionary")
                    for item in self.conf["items"]:
                        if item.get("name") in data:
                            item_name = item.get("name")
                            encoder = self.encoders[item_name]
                            res, ct , err = encoder.encode(data[item_name])
                            if err:
                                return res, "", ValueError(f"{item_name}: {str(err)}")
                            match item["type"]:
                                case "file" | "image" | "binary" | "audio" | "video":
                                    field = (f"{item_name}.bin", res, ct)
                                    form_data[item_name] = field
                                case _:
                                    form_data[item_name] = res
                    res, _ct, err = encoders.encode_multipart(form_data)
                    ct_items = _ct.split(":")
                    if len(ct_items) == 1:
                        contentType = ct_items[0].strip()
                    elif len(ct_items) > 1:
                        contentType = ct_items[1].strip()
                    else:
                        contentType = "multipart/form-data"
                case "json":
                    if len(self.json_images) > 0:
                        for entry, kind in zip(self.json_images, self.json_types):
                            tmp_data = get_by_path(data, entry)
                            if tmp_data and not isinstance(tmp_data, bytes):
                                # If it has already been encoded, skip
                                match kind:
                                    case "image":
                                        tmp_data, _ = encoders.encode_image(
                                            tmp_data)
                                    case _:
                                        tmp_data, _ = encoders.encode_file(
                                            tmp_data)
                                set_by_path(data, entry, tmp_data)
                    res, err = encoders.encode_json(data)
                    contentType = "application/json"
                case "matrix":
                    res, err = encoders.encode_json(data)
                    contentType = "application/json"
                case "file" | "binary" | "audio" | "video":
                    res, err = encoders.encode_file(data)
                    contentType = "application/octet-stream"
                case "image":
                    res, err = encoders.encode_image(data)
                    if not err:
                        contentType = f"image/{data.format.lower()}"
                    else:
                        contentType = "image/*"
                case "text" | "string":
                    res, err = encoders.encode_text(data)
                    contentType = "text/plain"

        except Exception as e:
            return res, "", ValueError(f"Error encoding data: {e}")
        return res, contentType, err


def get_by_path(root: dict, items: list[str]) -> Any:
    """Access a nested object in root by item sequence."""
    return reduce(operator.getitem, items, root)

def set_by_path(root: dict, items: list[str], value: Any):
    """Set a value in a nested object in root by item sequence."""
    get_by_path(root, items[:-1])[items[-1]] = value

def del_by_path(root: dict, items: list[str]):
    """Delete a key-value in a nested object in root by item sequence."""
    del get_by_path(root, items[:-1])[items[-1]]



"""
IA Parc Inference encoders
"""
from PIL.Image import Image
from io import BytesIO, StringIO

Error = ValueError | None

## Data encoders
def encode_file(file: BytesIO) -> tuple[bytes, Error]:
    """
    Encode file to bytes
    Arguments:
    file: BytesIO
    """
    if not file:
        return ''.encode(), ValueError("No data to encode")
    if not isinstance(file, BytesIO):
        return ''.encode(), ValueError("Data is not a file")
    try:
        data = file.read()
    except Exception as e:
        return ''.encode(), ValueError(f"Error encoding file: {e}")
    return data, None

def encode_image(img: Image) -> tuple[bytes, Error]:
    """
    Encode image to bytes
    Arguments:
    img: PIL Image
    """
    from PIL import Image
    data = ''.encode()
    if img is None:
        return data, ValueError("No data to encode")
    try:
        buf = BytesIO()
        if img.format == "" or img.format is None:
            img = img.convert("RGB")
            img.format = "JPEG"
        img.save(buf, format=img.format)
        data = buf.getvalue()
    except Exception as e:
        return data, ValueError(f"Error encoding image: {e}")
    return data, None

def encode_text(text: str) -> tuple[bytes, Error]:
    """
    Encode text to bytes
    Arguments:
    text: str
    """
    data = ''.encode()
    if not isinstance(text, str):
        return data, ValueError("Data is not a string")
    try:
        data = text.encode("utf-8")
    except Exception as e:
        return data, ValueError(f"Error encoding text: {e}")
    return data, None

def encode_json(in_data: dict) -> tuple[bytes, Error]:
    """
    Encode json to bytes
    Arguments:
    in_data: dict
    """
    data = ''.encode()
    from json_tricks import dumps
    if in_data is None:
        return data, ValueError("No data to encode")
    try:
        s_data = dumps(in_data)
        data = str(s_data).encode("utf-8")
    except Exception as e:
        return data, ValueError(f"Error encoding json: {e}")
    return data, None

def encode_numpy(in_data: dict) -> tuple[bytes, Error]:
    """
    Encode numpy to bytes
    Arguments:
    in_data: dict
    """
    return encode_json(in_data)

def encode_multipart(data: dict) -> tuple[bytes, str, Error]:
    """
    Encode multi-part data to bytes
    Arguments:
    data: dict
    """
    body = ''.encode()
    if data is None:
        return body, "", ValueError("No data to encode")
    try:
        from urllib3 import encode_multipart_formdata
        body, header = encode_multipart_formdata(data)
    except Exception as e:
        return body, "", ValueError(f"Error encoding multi-part data: {e}")
    return body, header, None




"""Config reader class"""
import logging.config
import os
import json
from json.decoder import JSONDecodeError
from yaml import Loader, dump, load


LOGGER = logging.getLogger("Config")

class Config(dict):
    """
    Config reader class
    """
    def __init__(self, data=None):
        super(Config, self).__init__()
        if data:
            if isinstance(data, dict):
                self.__update(data, {})
            elif isinstance(data, str):
                filename = os.path.basename(data)
                ext = os.path.splitext(filename)[1]
                self.__path = data
                self.__ext = ext
                if ext == "json":
                    self.__update(self.load_json(data), {})
                elif ext == "yaml" or ext == "yml":
                    self.__update(self.load_yaml(data), {})
                else:
                    try:
                        self.__update(self.load_json(data), {})
                    except (JSONDecodeError,TypeError):
                        self.__update(self.load_yaml(data), {})
            else:
                raise ValueError("Unknown data format")

    @staticmethod
    def dump_yaml(data, file_name):
        '''Dump data to yaml file'''
        to_dump = data.copy()
        del to_dump['_Config__path']
        del to_dump['_Config__ext']
        with open(f"{file_name}", "w", encoding="utf-8") as f:
            dump(to_dump, f)

    @staticmethod
    def dump_json(data, file_name):
        '''Dump data to json file'''
        to_dump = data.copy()
        del to_dump['_Config__path']
        del to_dump['_Config__ext']
        with open(f"{file_name}", "w", encoding="utf-8") as f:
            f.writelines(json.dumps(to_dump, indent=4))

    def save(self):
        '''Save config to file'''
        try:
            if self.__ext.lower() == ".json":
                self.save_to_json(self.__path)
            elif self.__ext.lower() == ".yaml" or self.__ext.lower() == ".yml":
                self.save_to_yaml(self.__path)
            else:
                LOGGER.error("Cannot save file, unknown extenstion %s", self.__ext)
        except Exception:
            LOGGER.error("Cannot save config", exc_info=True)

    def save_to_json(self, filename):
        '''Save config to json file'''
        self.dump_json(self, filename)

    def save_to_yaml(self, filename):
        '''Save config to yaml file'''
        self.dump_yaml(self, filename)

    @staticmethod
    def load_json(config):
        '''Load json file'''
        with open(config, "r", encoding="utf-8") as f:
            data = json.load(f)
        return data

    @staticmethod
    def load_yaml(config):
        '''Load yaml file'''
        with open(config, "r", encoding="utf-8") as f:
            data = load(f, Loader=Loader)
        return data

    def new(self, data):
        '''Create new config from data'''
        self.__update(data, {})

    def load(self, data, did):
        """load methode"""
        self.__update(data, did)

    def __update(self, data, did):
        dataid = id(data)
        did[dataid] = self
        for k in data:
            dkid = id(data[k])
            if dkid in did.keys():
                self[k] = did[dkid]
            elif isinstance(data[k], Config):
                self[k] = data[k]
            elif isinstance(data[k], dict):
                obj = Config()
                obj.load(data[k], did)
                self[k] = obj
                obj = None
            elif isinstance(data[k], list) or isinstance(data[k], tuple):
                self[k] = self._add_list(data[k], did)
            else:
                self[k] = data[k]

    def _add_list(self, data, did):
        lst = []
        for l in data:
            if isinstance(l, dict):
                obj = Config()
                obj.load(l, did)
                lst.append(obj)
                obj = None
            elif isinstance(l, list) or isinstance(l, tuple):
                lst.append(self._add_list(l, did))
            else:
                lst.append(l)
        if isinstance(data, tuple):
            lst = tuple(lst)
        return lst

    def __getattr__(self, key):
        return self.get(key, None)

    def __setattr__(self, key, value):
        if isinstance(value, dict):
            self[key] = Config(value)
        else:
            self[key] = value

    def has_key(self, k):
        """ returns True if key is present in the config"""
        if k in self.keys():
            return True
        else:
            return False

    def update(self, *args):
        for obj in args:
            for k in obj:
                if isinstance(obj[k], dict):
                    self[k] = Config(obj[k])
                else:
                    self[k] = obj[k]
        return self

    def merge(self, *args):
        """ merges the config with one or more configs"""
        for obj in args:
            for k in obj:
                if k in self.keys():
                    if isinstance(self[k], list) and isinstance(obj[k], list):
                        self[k] += obj[k]
                    elif isinstance(self[k], list):
                        self[k].append(obj[k])
                    elif isinstance(obj[k], list):
                        self[k] = [self[k]] + obj[k]
                    elif isinstance(self[k], Config) and isinstance(obj[k], Config):
                        self[k].merge(obj[k])
                    elif isinstance(self[k], Config) and isinstance(obj[k], dict):
                        self[k].merge(obj[k])
                    else:
                        self[k] = [self[k], obj[k]]
                else:
                    if isinstance(obj[k], dict):
                        self[k] = Config(obj[k])
                    else:
                        self[k] = obj[k]
        return self

    def replace_variables(self, variables):
        """ replaces all variables in the config with the given variables"""
        for k, obj in self.items():
            if isinstance(obj, Config):
                obj.replace_variables(variables)
            elif isinstance(obj, str):
                self[k] = obj.format(**variables)
            elif isinstance(obj, list) or isinstance(obj, tuple):
                self[k] = self.replace_in_list(obj, variables)

    def replace_in_list(self, obj, variables):
        """ replaces all variables in the list with the given variables"""
        for i, entry in enumerate(obj):
            if isinstance(entry, Config):
                entry.replace_variables(variables)
            elif isinstance(entry, str):
                obj[i] = entry.format(**variables)
            elif isinstance(obj, list) or isinstance(obj, tuple):
                self.replace_in_list(obj, variables)
        return obj

